#include "config.h"

#include <inttypes.h>
#include <unistd.h>
#include <time.h>

#include <cache/cache.h>
#include <vsb.h>

#include "vcc_tus_if.h"

#include "tus_file.h"
#include "tus_servers.h"
#include "tus_hdr.h"
#include "tus_response.h"
#include "tus_request.h"
#include "tus_stv.h"
#include "tus_hex.h"

/* ------------------------------------------------------------
 */
enum met_e {
	_INVALID = 0,
#define MET(x) x,
#include "tbl_method.h"
	_MET_MAX
};

static enum met_e
parse_met(const char *s)
{
#define MET(x) if (strcmp(s, #x) == 0) return (x);
#include "tbl_method.h"
	return (_INVALID);
}

// everything for a tus upload to the backend is complete
static VCL_BOOL
tus_request_complete(VRT_CTX, const struct VPFX(tus_server) *srv,
    struct tus_concat *c, struct tus_file_disk *fdisk)
{
	VCL_BLOB hash = NULL;
	struct vsb vsb[1];
	const char *slash;

	tus_body_assign(ctx->req, c);
	http_ForceField(ctx->http_req, HTTP_HDR_METHOD, "PUT");
	if (fdisk->type == TUS_SINGLE)
		hash = tus_concat_hash(ctx, srv, c);
	if (hash != NULL) {
		slash = strrchr(fdisk->url_path, '/');
		AN(slash);
		WS_VSB_new(vsb, ctx->ws);
		VSB_bcat(vsb, fdisk->url_path, (slash - fdisk->url_path) + 1);
		tus_vsbhex(vsb, hash);
		http_ForceField(ctx->http_req, HTTP_HDR_URL,
		    WS_VSB_finish(vsb, ctx->ws, NULL));
	} else {
		http_ForceField(ctx->http_req, HTTP_HDR_URL, fdisk->url_path);
	}
	http_Unset(ctx->http_req, H_Content_Length);
	http_PrintfHeader(ctx->http_req, "Content-Length: %zu",
	    fdisk->upload_length);

	// move to table?
	http_Unset(ctx->http_req, hdr_resum);
	http_Unset(ctx->http_req, hdr_vers);
	http_Unset(ctx->http_req, hdr_ext);
	http_Unset(ctx->http_req, hdr_chkalg);
	http_Unset(ctx->http_req, hdr_chksum);
	http_Unset(ctx->http_req, hdr_max);
	http_Unset(ctx->http_req, hdr_defer);
	http_Unset(ctx->http_req, hdr_meta);
	http_Unset(ctx->http_req, hdr_exp);
	http_Unset(ctx->http_req, hdr_loc);
	http_Unset(ctx->http_req, hdr_method);
	http_Unset(ctx->http_req, hdr_off);
	http_Unset(ctx->http_req, hdr_len);

	return (1);
}

struct test_meta {
	struct test_meta *prev;
	const char *k;
	size_t l;
};

static VCL_BOOL
tus_meta_key_unique(struct test_meta *this, size_t l)
{
	struct test_meta *prev = this;

	if (l == 0)
		return (0);
	if (this->l == 0)
		this->l = l;

	while ((prev = prev->prev) != NULL) {
		if (l != prev->l || strncmp(this->k, prev->k, l) != 0)
			continue;
		return (0);
	}

	return (1);
}

// from vmod_blob base64.c
#define ILL ((int8_t) 127)
#define PAD ((int8_t) 126)
const int8_t i64[256] = {
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL,  62, ILL, ILL, ILL,  63, /* +, -    */
	 52,  53,  54,  55,  56,  57,  58,  59, /* 0 - 7   */
	 60,  61, ILL, ILL, ILL, PAD, ILL, ILL, /* 8, 9, = */
	ILL,   0,   1,   2,   3,   4,   5,   6, /* A - G   */
	  7,   8,   9,  10,  11,  12,  13,  14, /* H - O   */
	 15,  16,  17,  18,  19,  20,  21,  22, /* P - W   */
	 23,  24,  25, ILL, ILL, ILL, ILL, ILL, /* X, Y, Z */
	ILL,  26,  27,  28,  29,  30,  31,  32, /* a - g   */
	 33,  34,  35,  36,  37,  38,  39,  40, /* h - o   */
	 41,  42,  43,  44,  45,  46,  47,  48, /* p - w   */
	 49,  50,  51, ILL, ILL, ILL, ILL, ILL, /* x, y, z */
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL,
	ILL, ILL, ILL, ILL, ILL, ILL, ILL, ILL
};

static VCL_BOOL
tus_metadata_validate(const char *s, struct test_meta *prev)
{
	struct test_meta this = {prev, s, 0};

	// key
	while (1) {
		if (*s != ',' && *s != ' ' && *s != '\0') {
			s++;
			continue;
		}

		if (tus_meta_key_unique(&this, s - this.k) == 0)
			return (0);

		if (*s == '\0')
			return (1);

		if (*s == ',' || (*s == ' ' && s[1] == ','))
			return (tus_metadata_validate(s + 1, &this));

		if (*s == ' ') {
			s++;
			break;
		}
	}
	while (*s != '\0') {
		if (*s == ',')
			return (tus_metadata_validate(s + 1, &this));
		if (i64[*s] == ILL)
			return (0);
		s++;
	}
	return (1);
}


VCL_BOOL
tus_request(VRT_CTX, struct VPFX(tus_server) *tussrv,
    struct tus_response *r, const char *url, const char *id)
{
	struct tus_file_core *fcore;
	struct tus_file_disk *fdisk = NULL;
	const char *p, *metadata = NULL, *concat = NULL;
	struct concat_embryo embryo;
	char *q;
	enum met_e m;
	int ct_ok, lock = EINVAL;
	uintmax_t cl, off = UINTMAX_MAX, len = UINTMAX_MAX;
	enum tus_f_type type = TUS_SINGLE;
	struct tus_concat *c;
	VCL_BLOB hash;

	AZ(r->status);
	AZ(r->fcore);

	if (http_GetHdr(ctx->http_req, hdr_method, &p)) {
		http_ForceField(ctx->http_req, HTTP_HDR_METHOD, p);
		http_Unset(ctx->http_req, hdr_method);
	}

	if (http_GetHdr(ctx->http_req, hdr_meta, &metadata) &&
	    tus_metadata_validate(metadata, NULL) == 0) {
		VSLb(ctx->vsl, SLT_Error, "%s: bad metadata format",
		    tus_server_name(tussrv));
		r->status = 400;
		return (0);
	}

	m = parse_met(http_GetMethod(ctx->http_req));

	(void) http_GetHdr(ctx->http_req, hdr_origin, &r->origin);

	if (m == _INVALID) {
		r->status = 405;
		return (0);
	}
	if (m == OPTIONS) {
		r->status = 204 + (s_OPTIONS * s_MULT);
		return (0);
	}

	if (http_GetHdr(ctx->http_req, hdr_concat, &concat)) {
		if (strcmp(concat, "partial") == 0) {
			type = TUS_PARTIAL;
		} else if (strncmp(concat, "final;", 6) == 0) {
			type = TUS_FINAL;
			concat += 6;
		} else {
			r->status = 400;
			return (0);
		}
	}

	if (m == POST && type == TUS_FINAL) {
		if (tus_file_final_concat(tussrv, &embryo, concat) == NULL) {
			r->status = 400;
			return (0);
		}
		hash = tus_concat_hash(ctx, embryo.srv, embryo.concat);
		if (hash != NULL) {
			id = tus_hex(ctx, hash);
			if (id == NULL) {
				r->status = 503;
				return (0);
			}
		}
	}

	AZ(pthread_mutex_lock(&tussrv->mtx));

	if (m == POST) {
		fcore = tus_file_new(ctx, tussrv, type, url, id, metadata);
	} else {
		fcore = tus_file_lookup(tussrv, url);
	}

	if (fcore != NULL)
		r->schemeauth = WS_Copy(ctx->ws, tussrv->schemeauth, -1);

	if (fcore != NULL)
		lock = pthread_mutex_trylock(&fcore->mtx);
	AZ(pthread_mutex_unlock(&tussrv->mtx));

	if (fcore != NULL) {
		fdisk = fcore->disk;
		AN(fdisk);
		type = fdisk->type;
	}

	// https://github.com/tus/tus-resumable-upload-protocol/issues/146
	if (fcore == NULL) {
		assert (lock == EINVAL);
	} else if (lock == EBUSY) {
		r->status = 423;
		return (0);
	} else {
		assert (lock == 0);
	}

	if (m == POST && type == TUS_FINAL) {
		if (fcore == NULL) {
			tus_file_final_abort(&embryo);
		} else {
			fcore = tus_file_final_birth(fcore, &embryo);
			r->status = 201;
		}
	}

	if (m == DELETE) {
		if (fcore == NULL) {
			r->status = 404;
		} else {
			AZ(pthread_mutex_lock(&tussrv->mtx));
			tus_file_del(&fcore);
			AZ(pthread_mutex_unlock(&tussrv->mtx));
			r->fcore = NULL;
			r->status = 204;
		}
		return (0);
	}

	if (type == TUS_REDIRECT) {
		r->fcore = fcore;
		r->status = 301;
		return (0);
	} else if (m == GET) {
		r->status = 400;
		return (0);
	}

	r->fcore = fcore;

	if (m == HEAD) {
		r->status = fcore ? 200 : 404;
		return (0);
	}

	if (fcore == NULL) {
		r->status = (m == POST) ? 409 : 404;
		return (0);
	}

	AN(fcore);
	AN(fdisk);

	ct_ok = http_GetHdr(ctx->http_req, H_Content_Type, &p) &&
	    (strcmp(p, "application/offset+octet-stream") == 0);

	if (http_GetHdr(ctx->http_req, hdr_off, &p)) {
		off = strtoumax(p, &q, 10);
		if (q == NULL || *q != '\0')
			off = UINTMAX_MAX;
	}

	if (http_GetHdr(ctx->http_req, hdr_len, &p)) {
		if (type == TUS_FINAL) {
			r->status = 400;
			return (0);
		}

		len = strtoumax(p, &q, 10);
		if (q == NULL || *q != '\0')
			len = UINTMAX_MAX;
	}
	if (http_GetHdr(ctx->http_req, hdr_defer, &p)) {
		if (len != UINTMAX_MAX || type == TUS_FINAL) {
			r->status = 400;
			return (0);
		}
		if (strcmp(p, "1") != 0) {
			r->status = 409; // or 400 ?
			return (0);
		}
	}

	if (len != UINTMAX_MAX) {
		if (fdisk->upload_length == -1)
			fdisk->upload_length = len;
		else if (len != fdisk->upload_length) {
			r->status = 409; // or 400 ?
			return (0);
		}
	}

	if (type == TUS_FINAL) {
		if (m != POST) {
			r->status = 403;
			r->fcore = NULL;
			return (0);
		}
		return (tus_request_complete(ctx, tussrv, fcore->ptr, fdisk));
	}

	assert (type != TUS_FINAL);

	switch (m) {
	case PATCH: {
		if (! ct_ok) {
			r->status = 415;
			return (0);
		}
		if (fcore == NULL) {
			r->status = 404;
			return (0);
		}
		if (off != fdisk->upload_offset) {
			r->status = 409;
			return (0);
		}
		r->status = tus_body_to_file(ctx, fcore);
		break;
	}
	case POST: {
		cl = 0;
		if (http_GetHdr(ctx->http_req, H_Content_Length, &p))
			cl = strtoumax(p, &q, 10);

		if (cl && ! ct_ok) {
			r->status = 415;
			return (0);
		}
		AZ(fdisk->upload_offset);
		if (cl == 0) {
			r->status = 201;
			break;
		}

		r->status = tus_body_to_file(ctx, fcore);
		if (r->status != 204)
			break;
		r->status = 201;
		break;
	}
	default:
		INCOMPL();
	}

	if (r->status == 413) {
		AZ(pthread_mutex_lock(&tussrv->mtx));
		tus_file_del(&fcore);
		AZ(pthread_mutex_unlock(&tussrv->mtx));
		r->fcore = NULL;
		return (0);
	}
	if (fdisk->upload_offset == fdisk->upload_length) {
		tus_file_mmap(fcore);
		AZ(pthread_cond_signal(&fcore->cond));
		if (type == TUS_PARTIAL)
			return (0);
		assert (type == TUS_SINGLE);
		c = tus_body_single(ctx->req, fcore);
		if (c == NULL) {
			VRT_fail(ctx, "not enough workspace");
			return (0);
		}
		return (tus_request_complete(ctx, tussrv, c, fdisk));
	}
	return (0);
}
